from __future__ import annotations

from ailment import Expr

from angr.engines.vex.claripy.ccall import data
from .rewriter_base import CCallRewriterBase


X86_CondTypes = data["X86"]["CondTypes"]
X86_OpTypes = data["X86"]["OpTypes"]
X86_CondBitMasks = data["X86"]["CondBitMasks"]
X86_CondBitOffsets = data["X86"]["CondBitOffsets"]


class X86CCallRewriter(CCallRewriterBase):
    """
    Implements VEX ccall rewriter for X86.
    """

    __slots__ = ()

    def _rewrite(self, ccall: Expr.VEXCCallExpression) -> Expr.Expression | None:
        if ccall.callee == "x86g_calculate_condition":
            cond = ccall.operands[0]
            op = ccall.operands[1]
            dep_1 = ccall.operands[2]
            dep_2 = ccall.operands[3]
            if isinstance(cond, Expr.Const) and isinstance(op, Expr.Const):
                cond_v = cond.value
                op_v = op.value
                if cond_v == X86_CondTypes["CondLE"]:  # noqa: SIM102
                    if op_v in {
                        X86_OpTypes["G_CC_OP_SUBB"],
                        X86_OpTypes["G_CC_OP_SUBW"],
                        X86_OpTypes["G_CC_OP_SUBL"],
                    }:
                        # dep_1 <=s dep_2
                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            X86_OpTypes["G_CC_OP_SUBB"],
                            X86_OpTypes["G_CC_OP_SUBW"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            X86_OpTypes["G_CC_OP_SUBB"],
                            X86_OpTypes["G_CC_OP_SUBW"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(ccall.idx, "CmpLE", (dep_1, dep_2), True, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
        return None

    @staticmethod
    def _fix_size(expr, op_v: int, type_8bit, type_16bit, tags):
        if op_v == type_8bit:
            bits = 8
        elif op_v == type_16bit:
            bits = 16
        else:
            bits = 32
        if bits < 32:
            if isinstance(expr, Expr.Const):
                return Expr.Const(expr.idx, None, expr.value & ((1 << bits) - 1), bits, **tags)
            return Expr.Convert(None, 32, bits, False, expr, **tags)
        return expr



from __future__ import annotations

import networkx

from ailment import AILBlockWalker, Block
from ailment.statement import ConditionalJump, Statement
from ailment.expression import Const, BinaryOp, VirtualVariable

from angr.analyses.decompiler.region_identifier import RegionIdentifier
from .optimization_pass import OptimizationPass, OptimizationPassStage


class ConstantCondition:
    """
    Describes an opportunity for replacing a vvar with a constant value.
    """

    def __init__(self, vvar_id: int, value: Const, block_addr: int, block_idx: int | None):
        self.vvar_id = vvar_id
        self.value = value
        self.block_addr = block_addr
        self.block_idx = block_idx

    def __repr__(self):
        return f"<ConstCond vvar_{self.vvar_id} == {self.value} since {self.block_addr:#x}-{self.block_idx}>"


class CCondPropBlockWalker(AILBlockWalker):
    """
    Block walker for ConditionConstantPropagation to replace vvars with constant values.
    """

    def __init__(self, vvar_id: int, const_value: Const):
        super().__init__()
        self._new_block: Block | None = None  # output
        self.vvar_id = vvar_id
        self.const_value = const_value

    def walk(self, block: Block):
        self._new_block = None
        super().walk(block)
        return self._new_block

    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Block):  # type: ignore
        r = super()._handle_stmt(stmt_idx, stmt, block)
        if r is not None:
            # replace the original statement
            if self._new_block is None:
                self._new_block = block.copy()
            self._new_block.statements[stmt_idx] = r

    def _handle_VirtualVariable(  # type: ignore
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement, block: Block | None
    ) -> Const | None:
        if expr.varid == self.vvar_id:
            return Const(expr.idx, None, self.const_value.value, self.const_value.bits, **expr.tags)
        return None


class ConditionConstantPropagation(OptimizationPass):
    """
    Reason about constant propagation opportunities from conditionals and propagate constants in the graph accordingly.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_SINGLE_BLOCK_SIMPLIFICATION
    NAME = "Propagate constants using information deduced from conditionals."
    DESCRIPTION = __doc__.strip()  # type: ignore

    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)
        self.analyze()

    def _check(self):
        cconds = self._find_const_conditions()
        if not cconds:
            return False, None
        return True, {"cconds": cconds}

    def _analyze(self, cache=None):
        if not cache or cache.get("cconds", None) is None:  # noqa: SIM108
            cconds = self._find_const_conditions()
        else:
            cconds = cache["cconds"]

        if not cconds:
            return

        # group cconds according to their sources
        cconds_by_src: dict[tuple[int, int | None], list[ConstantCondition]] = {}
        for ccond in cconds:
            src = ccond.block_addr, ccond.block_idx
            if src not in cconds_by_src:
                cconds_by_src[src] = []
            cconds_by_src[src].append(ccond)

        # calculate a dominance frontier for each block
        entry_node_addr, entry_node_idx = self.entry_node_addr
        entry_node = self._get_block(entry_node_addr, idx=entry_node_idx)
        df = networkx.algorithms.dominance_frontiers(self._graph, entry_node)

        for src, cconds in cconds_by_src.items():
            head_block = self._get_block(src[0], idx=src[1])
            if head_block is None:
                continue
            frontier = df.get(head_block)
            if frontier is None:
                continue
            graph_slice = RegionIdentifier.slice_graph(self._graph, head_block, frontier, include_frontier=False)
            for ccond in cconds:
                walker = CCondPropBlockWalker(ccond.vvar_id, ccond.value)
                for block in graph_slice:
                    new_block = walker.walk(block)
                    if new_block is not None:
                        self._update_block(block, new_block)

    def _find_const_conditions(self) -> list[ConstantCondition]:
        cconds = []

        for block in self._graph:
            if block.statements:
                last_stmt = block.statements[-1]
                if (
                    not isinstance(last_stmt, ConditionalJump)
                    or not isinstance(last_stmt.true_target, Const)
                    or not isinstance(last_stmt.false_target, Const)
                ):
                    continue

                if isinstance(last_stmt.condition, BinaryOp):
                    cond = last_stmt.condition
                    op = cond.op
                    op0, op1 = cond.operands
                    if isinstance(op0, Const):
                        op0, op1 = op1, op0
                    if isinstance(op0, VirtualVariable) and isinstance(op1, Const) and op1.is_int:
                        if op == "CmpEQ":
                            ccond = ConstantCondition(
                                op0.varid, op1, last_stmt.true_target.value, last_stmt.true_target_idx  # type: ignore
                            )
                            cconds.append(ccond)
                        elif op == "CmpNE":
                            ccond = ConstantCondition(
                                op0.varid, op1, last_stmt.false_target.value, last_stmt.false_target_idx  # type: ignore
                            )
                            cconds.append(ccond)

        return cconds
